@@ -311,7 +311,7 @@ de265_error de265_image::alloc_image(int w,int h, enum de265_chroma c,
     break;
   }
 
-  if (sps) {
+  if (chroma_format != de265_chroma_mono && sps) {
     assert(sps->SubWidthC  == SubWidthC);
     assert(sps->SubHeightC == SubHeightC);
   }
@@ -383,9 +383,15 @@ de265_error de265_image::alloc_image(int w,int h, enum de265_chroma c,
                                                               alloc_userdata);
 
     pixels_confwin[0] = pixels[0] + left*WinUnitX + top*WinUnitY*stride;
-    pixels_confwin[1] = pixels[1] + left + top*chroma_stride;
-    pixels_confwin[2] = pixels[2] + left + top*chroma_stride;
 
+    if (chroma_format != de265_chroma_mono) {
+      pixels_confwin[1] = pixels[1] + left + top*chroma_stride;
+      pixels_confwin[2] = pixels[2] + left + top*chroma_stride;
+    }
+    else {
+      pixels_confwin[1] = NULL;
+      pixels_confwin[2] = NULL;
+    }
 
     // check for memory shortage
 
