@@ -21,12 +21,18 @@
 #include "intrapred.h"
 #include "transform.h"
 #include "util.h"
+#include "encoder/encoder-types.h"
 #include <assert.h>
 
 
 #include <sys/types.h>
 #include <string.h>
 
+// Actually, the largest TB block can only be 32, but in some intra-pred-mode algorithms
+// (e.g. min-residual), we may call intra prediction on the maximum CTB size (64).
+static const int MAX_INTRA_PRED_BLOCK_SIZE = 64;
+
+
 
 #ifdef DE265_LOG_TRACE
 template <class pixel_t>
@@ -52,12 +58,58 @@ void print_border(pixel_t* data, uint8_t* available, int nT)
 #endif
 
 
-void fillIntraPredModeCandidates(int candModeList[3], int x,int y, int PUidx,
+
+void fillIntraPredModeCandidates(enum IntraPredMode candModeList[3],
+                                 enum IntraPredMode candIntraPredModeA,
+                                 enum IntraPredMode candIntraPredModeB)
+{
+  // build candidate list
+
+  if (candIntraPredModeA == candIntraPredModeB) {
+    if (candIntraPredModeA < 2) {
+      candModeList[0] = INTRA_PLANAR;
+      candModeList[1] = INTRA_DC;
+      candModeList[2] = INTRA_ANGULAR_26;
+    }
+    else {
+      candModeList[0] = candIntraPredModeA;
+      candModeList[1] = (enum IntraPredMode)(2 + ((candIntraPredModeA-2 -1 +32) % 32));
+      candModeList[2] = (enum IntraPredMode)(2 + ((candIntraPredModeA-2 +1    ) % 32));
+    }
+  }
+  else {
+    candModeList[0] = candIntraPredModeA;
+    candModeList[1] = candIntraPredModeB;
+
+    if (candIntraPredModeA != INTRA_PLANAR &&
+        candIntraPredModeB != INTRA_PLANAR) {
+      candModeList[2] = INTRA_PLANAR;
+    }
+    else if (candIntraPredModeA != INTRA_DC &&
+             candIntraPredModeB != INTRA_DC) {
+      candModeList[2] = INTRA_DC;
+    }
+    else {
+      candModeList[2] = INTRA_ANGULAR_26;
+    }
+  }
+
+  /*
+    printf("candModeList: %d %d %d\n",
+    candModeList[0],
+    candModeList[1],
+    candModeList[2]
+    );
+  */
+}
+
+
+void fillIntraPredModeCandidates(enum IntraPredMode candModeList[3], int x,int y, int PUidx,
                                  bool availableA, // left
                                  bool availableB, // top
                                  const de265_image* img)
 {
-  const seq_parameter_set* sps = &img->sps;
+  const seq_parameter_set* sps = &img->get_sps();
 
   // block on left side
 
@@ -90,53 +142,85 @@ void fillIntraPredModeCandidates(int candModeList[3], int x,int y, int PUidx,
   }
 
 
-  // build candidate list
+  logtrace(LogSlice,"%d;%d candA:%d / candB:%d\n", x,y,
+           availableA ? candIntraPredModeA : -999,
+           availableB ? candIntraPredModeB : -999);
 
-  logtrace(LogSlice,"%d;%d availableA:%d candA:%d & availableB:%d candB:%d\n", x,y,
-           availableA, candIntraPredModeA,
-           availableB, candIntraPredModeB);
 
-  if (candIntraPredModeA == candIntraPredModeB) {
-    if (candIntraPredModeA < 2) {
-      candModeList[0] = INTRA_PLANAR;
-      candModeList[1] = INTRA_DC;
-      candModeList[2] = INTRA_ANGULAR_26;
-    }
-    else {
-      candModeList[0] = candIntraPredModeA;
-      candModeList[1] = 2 + ((candIntraPredModeA-2 -1 +32) % 32);
-      candModeList[2] = 2 + ((candIntraPredModeA-2 +1    ) % 32);
-    }
+  fillIntraPredModeCandidates(candModeList,
+                              candIntraPredModeA,
+                              candIntraPredModeB);
+}
+
+
+void fillIntraPredModeCandidates(enum IntraPredMode candModeList[3],
+                                 int x,int y,
+                                 bool availableA, // left
+                                 bool availableB, // top
+                                 const CTBTreeMatrix& ctbs,
+                                 const seq_parameter_set* sps)
+{
+
+  // block on left side
+
+  enum IntraPredMode candIntraPredModeA, candIntraPredModeB;
+
+  if (availableA==false) {
+    candIntraPredModeA=INTRA_DC;
   }
   else {
-    candModeList[0] = candIntraPredModeA;
-    candModeList[1] = candIntraPredModeB;
+    const enc_cb* cbL = ctbs.getCB(x-1,y);
+    assert(cbL != NULL);
 
-    if (candIntraPredModeA != INTRA_PLANAR &&
-        candIntraPredModeB != INTRA_PLANAR) {
-      candModeList[2] = INTRA_PLANAR;
-    }
-    else if (candIntraPredModeA != INTRA_DC &&
-             candIntraPredModeB != INTRA_DC) {
-      candModeList[2] = INTRA_DC;
+    if (cbL->PredMode != MODE_INTRA ||
+        cbL->pcm_flag) {
+      candIntraPredModeA=INTRA_DC;
     }
     else {
-      candModeList[2] = INTRA_ANGULAR_26;
+      const enc_tb* tbL = cbL->getTB(x-1,y);
+      assert(tbL);
+      candIntraPredModeA = tbL->intra_mode;
     }
   }
 
-  /*
-    printf("candModeList: %d %d %d\n",
-    candModeList[0],
-    candModeList[1],
-    candModeList[2]
-    );
-  */
+  // block above
+
+  if (availableB==false) {
+    candIntraPredModeB=INTRA_DC;
+  }
+  else {
+    const enc_cb* cbA = ctbs.getCB(x,y-1);
+    assert(cbA != NULL);
+
+    if (cbA->PredMode != MODE_INTRA ||
+        cbA->pcm_flag) {
+      candIntraPredModeB=INTRA_DC;
+    }
+    else if (y-1 < ((y >> sps->Log2CtbSizeY) << sps->Log2CtbSizeY)) {
+      candIntraPredModeB=INTRA_DC;
+    }
+    else {
+      const enc_tb* tbA = cbA->getTB(x,y-1);
+      assert(tbA);
+
+      candIntraPredModeB = tbA->intra_mode;
+    }
+  }
+
+
+  logtrace(LogSlice,"%d;%d candA:%d / candB:%d\n", x,y,
+           availableA ? candIntraPredModeA : -999,
+           availableB ? candIntraPredModeB : -999);
+
+
+  fillIntraPredModeCandidates(candModeList,
+                              candIntraPredModeA,
+                              candIntraPredModeB);
 }
 
 
 int find_intra_pred_mode(enum IntraPredMode mode,
-                         int candModeList[3])
+                         enum IntraPredMode candModeList[3])
 {
   // check whether the mode is in the candidate list
 
@@ -257,29 +341,107 @@ enum IntraPredMode lumaPredMode_to_chromaPredMode(enum IntraPredMode luma,
 }
 
 
-// (8.4.4.2.2)
 template <class pixel_t>
-void fill_border_samples(de265_image* img,
-                         int xB,int yB,  // in component specific resolution
-                         int nT, int cIdx,
-                         pixel_t* out_border)
+struct intra_border_computer
 {
-  const seq_parameter_set* sps = &img->sps;
-  const pic_parameter_set* pps = &img->pps;
+  pixel_t* out_border;
 
-  uint8_t available_data[2*64 + 1];
-  uint8_t* available = &available_data[64];
+  const de265_image* img;
+  int nT;
+  int cIdx;
 
-  pixel_t* image;
-  int stride;
-  image  = (pixel_t*)img->get_image_plane(cIdx);
-  stride = img->get_image_stride(cIdx);
+  int xB,yB;
 
-  const int SubWidth  = (cIdx==0) ? 1 : sps->SubWidthC;
-  const int SubHeight = (cIdx==0) ? 1 : sps->SubHeightC;
+  const seq_parameter_set* sps;
+  const pic_parameter_set* pps;
+
+  uint8_t available_data[4*MAX_INTRA_PRED_BLOCK_SIZE + 1];
+  uint8_t* available;
+
+  int SubWidth;
+  int SubHeight;
+
+  bool availableLeft;    // is CTB at left side available?
+  bool availableTop;     // is CTB at top side available?
+  bool availableTopRight; // is CTB at top-right side available?
+  bool availableTopLeft;  // if CTB at top-left pixel available?
+
+  int nBottom;
+  int nRight;
+  int nAvail;
+  pixel_t firstValue;
+
+  void init(pixel_t* _out_border,
+            const de265_image* _img, int _nT, int _cIdx, int _xB, int _yB) {
+    img=_img; nT=_nT; cIdx=_cIdx;
+    out_border=_out_border; xB=_xB; yB=_yB;
+
+    assert(nT <= MAX_INTRA_PRED_BLOCK_SIZE);
+
+    availableLeft=true;
+    availableTop=true;
+    availableTopRight=true;
+    availableTopLeft=true;
+  }
+  void preproc();
+  void fill_from_image();
+  void fill_from_ctbtree(const enc_tb* tb,
+                         const CTBTreeMatrix& ctbs);
+  void reference_sample_substitution();
+};
+
+
+template <class pixel_t>
+void intra_border_computer<pixel_t>::reference_sample_substitution()
+{
+  // reference sample substitution
 
   const int bit_depth = img->get_bit_depth(cIdx);
 
+  if (nAvail!=4*nT+1) {
+    if (nAvail==0) {
+      if (sizeof(pixel_t)==1) {
+        memset(out_border-2*nT, 1<<(bit_depth-1), 4*nT+1);
+      }
+      else {
+        for (int i = -2*nT; i <= 2*nT ; i++) {
+          out_border[i] = 1<<(bit_depth-1);
+        }
+      }
+    }
+    else {
+      if (!available[-2*nT]) {
+        out_border[-2*nT] = firstValue;
+      }
+
+      for (int i=-2*nT+1; i<=2*nT; i++)
+        if (!available[i]) {
+          out_border[i]=out_border[i-1];
+        }
+    }
+  }
+
+  logtrace(LogIntraPred,"availableN: ");
+  print_border(available,NULL,nT);
+  logtrace(LogIntraPred,"\n");
+
+  logtrace(LogIntraPred,"output:     ");
+  print_border(out_border,NULL,nT);
+  logtrace(LogIntraPred,"\n");
+}
+
+
+
+
+template <class pixel_t>
+void intra_border_computer<pixel_t>::preproc()
+{
+  sps = &img->get_sps();
+  pps = &img->get_pps();
+
+  SubWidth  = (cIdx==0) ? 1 : sps->SubWidthC;
+  SubHeight = (cIdx==0) ? 1 : sps->SubHeightC;
+
   // --- check for CTB boundaries ---
 
   int xBLuma = xB * SubWidth;
@@ -288,11 +450,6 @@ void fill_border_samples(de265_image* img,
   int log2CtbSize = sps->Log2CtbSizeY;
   int picWidthInCtbs = sps->PicWidthInCtbsY;
 
-  bool availableLeft=true;    // is CTB at left side available?
-  bool availableTop=true;     // is CTB at top side available?
-  bool availableTopRight=true; // is CTB at top-right side available?
-  bool availableTopLeft=true;  // if CTB at top-left pixel available?
-
 
   //printf("xB/yB: %d %d\n",xB,yB);
 
@@ -350,186 +507,352 @@ void fill_border_samples(de265_image* img,
   if (topleftCTBSlice !=currCTBSlice||topleftCTBTileID!=currCTBTileID ) availableTopLeft = false;
   if (toprightCTBSlice!=currCTBSlice||toprightCTBTileID!=currCTBTileID) availableTopRight= false;
 
+
+  // number of pixels that are in the valid image area to the right and to the bottom
+
+  nBottom = sps->pic_height_in_luma_samples - yB*SubHeight;
+  nBottom=(nBottom+SubHeight-1)/SubHeight;
+  if (nBottom>2*nT) nBottom=2*nT;
+
+  nRight  = sps->pic_width_in_luma_samples  - xB*SubWidth;
+  nRight =(nRight +SubWidth-1)/SubWidth;
+  if (nRight >2*nT) nRight=2*nT;
+
+  nAvail=0;
+
+  available = &available_data[2*MAX_INTRA_PRED_BLOCK_SIZE];
+
+  memset(available-2*nT, 0, 4*nT+1);
+}
+
+
+
+// (8.4.4.2.2)
+template <class pixel_t>
+void fill_border_samples(de265_image* img,
+                         int xB,int yB,  // in component specific resolution
+                         int nT, int cIdx,
+                         pixel_t* out_border)
+{
+  intra_border_computer<pixel_t> c;
+  c.init(out_border, img, nT, cIdx, xB, yB);
+  c.preproc();
+  c.fill_from_image();
+  c.reference_sample_substitution();
+}
+
+
+// (8.4.4.2.2)
+template <class pixel_t>
+void fill_border_samples_from_tree(const de265_image* img,
+                                   const enc_tb* tb,
+                                   const CTBTreeMatrix& ctbs,
+                                   int cIdx,
+                                   pixel_t* out_border)
+{
+  intra_border_computer<pixel_t> c;
+
+  // xB,yB in component specific resolution
+  int xB,yB;
+  int nT = 1<<tb->log2Size;
+
+  xB = tb->x;
+  yB = tb->y;
+
+  if (img->get_sps().chroma_format_idc == CHROMA_444) {
+  }
+  else if (cIdx > 0) {
+    // TODO: proper chroma handling
+    xB >>= 1;
+    yB >>= 1;
+    nT >>= 1;
+
+    if (tb->log2Size==2) {
+      xB = tb->parent->x >> 1;
+      yB = tb->parent->y >> 1;
+      nT = 4;
+    }
+  }
+
+  c.init(out_border, img, nT, cIdx, xB, yB);
+  c.preproc();
+  c.fill_from_ctbtree(tb, ctbs);
+  c.reference_sample_substitution();
+}
+
+
+template <class pixel_t>
+void intra_border_computer<pixel_t>::fill_from_image()
+{
+  assert(nT<=32);
+
+  pixel_t* image;
+  int stride;
+  image  = (pixel_t*)img->get_image_plane(cIdx);
+  stride = img->get_image_stride(cIdx);
+
+  int xBLuma = xB * SubWidth;
+  int yBLuma = yB * SubHeight;
+
   int currBlockAddr = pps->MinTbAddrZS[ (xBLuma>>sps->Log2MinTrafoSize) +
                                         (yBLuma>>sps->Log2MinTrafoSize) * sps->PicWidthInTbsY ];
 
 
-  // number of pixels that are in the valid image area to the right and to the bottom
+  // copy pixels at left column
 
-  int nBottom = sps->pic_height_in_luma_samples - yB*SubHeight;
-  nBottom=(nBottom+SubHeight-1)/SubHeight;
-  if (nBottom>2*nT) nBottom=2*nT;
-
-  int nRight  = sps->pic_width_in_luma_samples  - xB*SubWidth;
-  nRight =(nRight +SubWidth-1)/SubWidth;
-  if (nRight >2*nT) nRight=2*nT;
-
-  int nAvail=0;
-
-  pixel_t firstValue;
-
-  memset(available-2*nT, 0, 4*nT+1);
-
-  {
-    // copy pixels at left column
-
-    for (int y=nBottom-1 ; y>=0 ; y-=4)
-      if (availableLeft)
-        {
-          int NBlockAddr = pps->MinTbAddrZS[ (((xB-1)*SubWidth )>>sps->Log2MinTrafoSize) +
-                                             (((yB+y)*SubHeight)>>sps->Log2MinTrafoSize)
-                                             * sps->PicWidthInTbsY ];
-
-          bool availableN = NBlockAddr <= currBlockAddr;
-
-          if (pps->constrained_intra_pred_flag) {
-            if (img->get_pred_mode((xB-1)*SubWidth,(yB+y)*SubHeight)!=MODE_INTRA)
-              availableN = false;
-          }
-
-          if (availableN) {
-            if (!nAvail) firstValue = image[xB-1 + (yB+y)*stride];
-
-            for (int i=0;i<4;i++) {
-              available[-y+i-1] = availableN;
-              out_border[-y+i-1] = image[xB-1 + (yB+y-i)*stride];
-            }
-
-            nAvail+=4;
-          }
-        }
-
-    // copy pixel at top-left position
-
-    if (availableTopLeft)
+  for (int y=nBottom-1 ; y>=0 ; y-=4)
+    if (availableLeft)
       {
         int NBlockAddr = pps->MinTbAddrZS[ (((xB-1)*SubWidth )>>sps->Log2MinTrafoSize) +
+                                           (((yB+y)*SubHeight)>>sps->Log2MinTrafoSize)
+                                           * sps->PicWidthInTbsY ];
+
+        bool availableN = NBlockAddr <= currBlockAddr;
+
+        if (pps->constrained_intra_pred_flag) {
+          if (img->get_pred_mode((xB-1)*SubWidth,(yB+y)*SubHeight)!=MODE_INTRA)
+            availableN = false;
+        }
+
+        if (availableN) {
+          if (!nAvail) firstValue = image[xB-1 + (yB+y)*stride];
+
+          for (int i=0;i<4;i++) {
+            available[-y+i-1] = availableN;
+            out_border[-y+i-1] = image[xB-1 + (yB+y-i)*stride];
+          }
+
+          nAvail+=4;
+        }
+      }
+
+  // copy pixel at top-left position
+
+  if (availableTopLeft)
+    {
+      int NBlockAddr = pps->MinTbAddrZS[ (((xB-1)*SubWidth )>>sps->Log2MinTrafoSize) +
+                                         (((yB-1)*SubHeight)>>sps->Log2MinTrafoSize)
+                                         * sps->PicWidthInTbsY ];
+
+      bool availableN = NBlockAddr <= currBlockAddr;
+
+      if (pps->constrained_intra_pred_flag) {
+        if (img->get_pred_mode((xB-1)*SubWidth,(yB-1)*SubHeight)!=MODE_INTRA) {
+          availableN = false;
+        }
+      }
+
+      if (availableN) {
+        if (!nAvail) firstValue = image[xB-1 + (yB-1)*stride];
+
+        out_border[0] = image[xB-1 + (yB-1)*stride];
+        available[0] = availableN;
+        nAvail++;
+      }
+    }
+
+  // copy pixels at top row
+
+  for (int x=0 ; x<nRight ; x+=4) {
+    bool borderAvailable;
+    if (x<nT) borderAvailable=availableTop;
+    else      borderAvailable=availableTopRight;
+
+    if (borderAvailable)
+      {
+        int NBlockAddr = pps->MinTbAddrZS[ (((xB+x)*SubWidth )>>sps->Log2MinTrafoSize) +
                                            (((yB-1)*SubHeight)>>sps->Log2MinTrafoSize)
                                            * sps->PicWidthInTbsY ];
 
         bool availableN = NBlockAddr <= currBlockAddr;
 
         if (pps->constrained_intra_pred_flag) {
-          if (img->get_pred_mode((xB-1)*SubWidth,(yB-1)*SubHeight)!=MODE_INTRA) {
+          if (img->get_pred_mode((xB+x)*SubWidth,(yB-1)*SubHeight)!=MODE_INTRA) {
             availableN = false;
           }
         }
 
+
         if (availableN) {
-          if (!nAvail) firstValue = image[xB-1 + (yB-1)*stride];
+          if (!nAvail) firstValue = image[xB+x + (yB-1)*stride];
 
-          out_border[0] = image[xB-1 + (yB-1)*stride];
-          available[0] = availableN;
-          nAvail++;
+          for (int i=0;i<4;i++) {
+            out_border[x+i+1] = image[xB+x+i + (yB-1)*stride];
+            available[x+i+1] = availableN;
+          }
+
+          nAvail+=4;
+        }
+      }
+  }
+}
+
+
+template <class pixel_t>
+void intra_border_computer<pixel_t>::fill_from_ctbtree(const enc_tb* blkTb,
+                                                       const CTBTreeMatrix& ctbs)
+{
+  int xBLuma = xB * SubWidth;
+  int yBLuma = yB * SubHeight;
+
+  int currBlockAddr = pps->MinTbAddrZS[ (xBLuma>>sps->Log2MinTrafoSize) +
+                                        (yBLuma>>sps->Log2MinTrafoSize) * sps->PicWidthInTbsY ];
+
+
+  // copy pixels at left column
+
+  for (int y=nBottom-1 ; y>=0 ; y-=4)
+    if (availableLeft)
+      {
+        int NBlockAddr = pps->MinTbAddrZS[ (((xB-1)*SubWidth )>>sps->Log2MinTrafoSize) +
+                                           (((yB+y)*SubHeight)>>sps->Log2MinTrafoSize)
+                                           * sps->PicWidthInTbsY ];
+
+        bool availableN = NBlockAddr <= currBlockAddr;
+
+        int xN = xB-1;
+        int yN = yB+y;
+
+        const enc_cb* cb = ctbs.getCB(xN*SubWidth, yN*SubHeight);
+
+        if (pps->constrained_intra_pred_flag) {
+          if (cb->PredMode != MODE_INTRA)
+            availableN = false;
+        }
+
+        if (availableN) {
+          PixelAccessor pa = cb->transform_tree->getPixels(xN,yN, cIdx, *sps);
+
+          if (!nAvail) firstValue = pa[yB+y][xB-1];
+
+          for (int i=0;i<4;i++) {
+            available[-y+i-1] = availableN;
+            out_border[-y+i-1] = pa[yB+y-i][xB-1];
+          }
+
+          nAvail+=4;
         }
       }
 
-    // copy pixels at top row
+  // copy pixel at top-left position
 
-    for (int x=0 ; x<nRight ; x+=4) {
-      bool borderAvailable;
-      if (x<nT) borderAvailable=availableTop;
-      else      borderAvailable=availableTopRight;
+  if (availableTopLeft)
+    {
+      int NBlockAddr = pps->MinTbAddrZS[ (((xB-1)*SubWidth )>>sps->Log2MinTrafoSize) +
+                                         (((yB-1)*SubHeight)>>sps->Log2MinTrafoSize)
+                                         * sps->PicWidthInTbsY ];
 
-      if (borderAvailable)
-        {
-          int NBlockAddr = pps->MinTbAddrZS[ (((xB+x)*SubWidth )>>sps->Log2MinTrafoSize) +
-                                             (((yB-1)*SubHeight)>>sps->Log2MinTrafoSize)
-                                             * sps->PicWidthInTbsY ];
+      bool availableN = NBlockAddr <= currBlockAddr;
 
-          bool availableN = NBlockAddr <= currBlockAddr;
+      int xN = xB-1;
+      int yN = yB-1;
 
-          if (pps->constrained_intra_pred_flag) {
-            if (img->get_pred_mode((xB+x)*SubWidth,(yB-1)*SubHeight)!=MODE_INTRA) {
-              availableN = false;
-            }
-          }
+      const enc_cb* cb = ctbs.getCB(xN*SubWidth, yN*SubHeight);
 
-
-          if (availableN) {
-            if (!nAvail) firstValue = image[xB+x + (yB-1)*stride];
-
-            for (int i=0;i<4;i++) {
-              out_border[x+i+1] = image[xB+x+i + (yB-1)*stride];
-              available[x+i+1] = availableN;
-            }
-
-            nAvail+=4;
-          }
-        }
-    }
-
-
-    // reference sample substitution
-
-    if (nAvail!=4*nT+1) {
-      if (nAvail==0) {
-        if (sizeof(pixel_t)==1) {
-          memset(out_border-2*nT, 1<<(bit_depth-1), 4*nT+1);
-        }
-        else {
-          for (int i = -2*nT; i <= 2*nT ; i++) {
-            out_border[i] = 1<<(bit_depth-1);
-          }
+      if (pps->constrained_intra_pred_flag) {
+        if (cb->PredMode!=MODE_INTRA) {
+          availableN = false;
         }
       }
-      else {
-        if (!available[-2*nT]) {
-          out_border[-2*nT] = firstValue;
-        }
 
-        for (int i=-2*nT+1; i<=2*nT; i++)
-          if (!available[i]) {
-            out_border[i]=out_border[i-1];
-          }
+      if (availableN) {
+        PixelAccessor pa = cb->transform_tree->getPixels(xN,yN, cIdx, *sps);
+
+        out_border[0] = pa[yB-1][xB-1];
+        available[0] = availableN;
+
+        if (!nAvail) firstValue = out_border[0];
+        nAvail++;
       }
     }
 
-    logtrace(LogIntraPred,"availableN: ");
-    print_border(available,NULL,nT);
-    logtrace(LogIntraPred,"\n");
 
-    logtrace(LogIntraPred,"output:     ");
-    print_border(out_border,NULL,nT);
-    logtrace(LogIntraPred,"\n");
+  // copy pixels at top row
+
+  for (int x=0 ; x<nRight ; x+=4) {
+    bool borderAvailable;
+    if (x<nT) borderAvailable=availableTop;
+    else      borderAvailable=availableTopRight;
+
+    if (borderAvailable)
+      {
+        int NBlockAddr = pps->MinTbAddrZS[ (((xB+x)*SubWidth )>>sps->Log2MinTrafoSize) +
+                                           (((yB-1)*SubHeight)>>sps->Log2MinTrafoSize)
+                                           * sps->PicWidthInTbsY ];
+
+        bool availableN = NBlockAddr <= currBlockAddr;
+
+        int xN = xB+x;
+        int yN = yB-1;
+
+        const enc_cb* cb = ctbs.getCB(xN*SubWidth, yN*SubHeight);
+
+        if (pps->constrained_intra_pred_flag) {
+          if (cb->PredMode!=MODE_INTRA) {
+            availableN = false;
+          }
+        }
+
+
+        if (availableN) {
+          PixelAccessor pa = cb->transform_tree->getPixels(xN,yN, cIdx, *sps);
+
+          if (!nAvail) firstValue = pa[yB-1][xB+x];
+
+          for (int i=0;i<4;i++) {
+            out_border[x+i+1] = pa[yB-1][xB+x+i];
+            available[x+i+1] = availableN;
+          }
+
+          nAvail+=4;
+        }
+      }
   }
 }
 
 
 // (8.4.4.2.3)
 template <class pixel_t>
-void intra_prediction_sample_filtering(de265_image* img,
+void intra_prediction_sample_filtering(const seq_parameter_set& sps,
                                        pixel_t* p,
                                        int nT, int cIdx,
                                        enum IntraPredMode intraPredMode)
 {
   int filterFlag;
 
+  //printf("filtering, mode: %d\n",intraPredMode);
+
   if (intraPredMode==INTRA_DC || nT==4) {
     filterFlag = 0;
   } else {
     // int-cast below prevents a typing problem that leads to wrong results when abs_value is a macro
     int minDistVerHor = libde265_min( abs_value((int)intraPredMode-26),
                                       abs_value((int)intraPredMode-10) );
+
+    //printf("mindist: %d\n",minDistVerHor);
+
     switch (nT) {
     case 8:  filterFlag = (minDistVerHor>7) ? 1 : 0; break;
     case 16: filterFlag = (minDistVerHor>1) ? 1 : 0; break;
     case 32: filterFlag = (minDistVerHor>0) ? 1 : 0; break;
+      // there is no official 64x64 TB block, but we call this for some intra-pred mode algorithms
+      // on the whole CB (2Nx2N mode for the whole CTB)
+    case 64: filterFlag = 0; break;
     default: filterFlag = -1; assert(false); break; // should never happen
     }
   }
 
 
   if (filterFlag) {
-    int biIntFlag = (img->sps.strong_intra_smoothing_enable_flag &&
+    int biIntFlag = (sps.strong_intra_smoothing_enable_flag &&
                      cIdx==0 &&
                      nT==32 &&
-                     abs_value(p[0]+p[ 64]-2*p[ 32]) < (1<<(img->sps.bit_depth_luma-5)) &&
-                     abs_value(p[0]+p[-64]-2*p[-32]) < (1<<(img->sps.bit_depth_luma-5)))
+                     abs_value(p[0]+p[ 64]-2*p[ 32]) < (1<<(sps.bit_depth_luma-5)) &&
+                     abs_value(p[0]+p[-64]-2*p[-32]) < (1<<(sps.bit_depth_luma-5)))
       ? 1 : 0;
 
-    pixel_t  pF_mem[2*64+1];
-    pixel_t* pF = &pF_mem[64];
+    pixel_t  pF_mem[4*32+1];
+    pixel_t* pF = &pF_mem[2*32];
 
     if (biIntFlag) {
       pF[-2*nT] = p[-2*nT];
@@ -577,31 +900,21 @@ static const int invAngle_table[25-10] =
 
 // (8.4.4.2.6)
 template <class pixel_t>
-void intra_prediction_angular(de265_image* img,
+void intra_prediction_angular(pixel_t* dst, int dstStride,
+                              int bit_depth, bool disableIntraBoundaryFilter,
                               int xB0,int yB0,
                               enum IntraPredMode intraPredMode,
                               int nT,int cIdx,
                               pixel_t* border)
 {
-  pixel_t  ref_mem[2*64+1];
-  pixel_t* ref=&ref_mem[64];
-
-  pixel_t* pred;
-  int      stride;
-  pred   = img->get_image_plane_at_pos_NEW<pixel_t>(cIdx,xB0,yB0);
-  stride = img->get_image_stride(cIdx);
-
-  int bit_depth = img->get_bit_depth(cIdx);
+  pixel_t  ref_mem[4*MAX_INTRA_PRED_BLOCK_SIZE+1]; // TODO: what is the required range here ?
+  pixel_t* ref=&ref_mem[2*MAX_INTRA_PRED_BLOCK_SIZE];
 
   assert(intraPredMode<35);
   assert(intraPredMode>=2);
 
   int intraPredAngle = intraPredAngle_table[intraPredMode];
 
-  bool disableIntraBoundaryFilter =
-    (img->sps.range_extension.implicit_rdpcm_enabled_flag &&
-     img->get_cu_transquant_bypass(xB0,yB0));
-
   if (intraPredMode >= 18) {
 
     for (int x=0;x<=nT;x++)
@@ -628,15 +941,15 @@ void intra_prediction_angular(de265_image* img,
           int iFact= ((y+1)*intraPredAngle)&31;
 
           if (iFact != 0) {
-            pred[x+y*stride] = ((32-iFact)*ref[x+iIdx+1] + iFact*ref[x+iIdx+2] + 16)>>5;
+            dst[x+y*dstStride] = ((32-iFact)*ref[x+iIdx+1] + iFact*ref[x+iIdx+2] + 16)>>5;
           } else {
-            pred[x+y*stride] = ref[x+iIdx+1];
+            dst[x+y*dstStride] = ref[x+iIdx+1];
           }
         }
 
     if (intraPredMode==26 && cIdx==0 && nT<32 && !disableIntraBoundaryFilter) {
       for (int y=0;y<nT;y++) {
-        pred[0+y*stride] = Clip_BitDepth(border[1] + ((border[-1-y] - border[0])>>1), bit_depth);
+        dst[0+y*dstStride] = Clip_BitDepth(border[1] + ((border[-1-y] - border[0])>>1), bit_depth);
       }
     }
   }
@@ -666,15 +979,15 @@ void intra_prediction_angular(de265_image* img,
           int iFact= ((x+1)*intraPredAngle)&31;  // DIFF (x<->y)
 
           if (iFact != 0) {
-            pred[x+y*stride] = ((32-iFact)*ref[y+iIdx+1] + iFact*ref[y+iIdx+2] + 16)>>5; // DIFF (x<->y)
+            dst[x+y*dstStride] = ((32-iFact)*ref[y+iIdx+1] + iFact*ref[y+iIdx+2] + 16)>>5; // DIFF (x<->y)
           } else {
-            pred[x+y*stride] = ref[y+iIdx+1]; // DIFF (x<->y)
+            dst[x+y*dstStride] = ref[y+iIdx+1]; // DIFF (x<->y)
           }
         }
 
     if (intraPredMode==10 && cIdx==0 && nT<32 && !disableIntraBoundaryFilter) {  // DIFF 26->10
       for (int x=0;x<nT;x++) { // DIFF (x<->y)
-        pred[x] = Clip_BitDepth(border[-1] + ((border[1+x] - border[0])>>1), bit_depth); // DIFF (x<->y && neg)
+        dst[x] = Clip_BitDepth(border[-1] + ((border[1+x] - border[0])>>1), bit_depth); // DIFF (x<->y && neg)
       }
     }
   }
@@ -685,7 +998,7 @@ void intra_prediction_angular(de265_image* img,
   for (int y=0;y<nT;y++)
     {
       for (int x=0;x<nT;x++)
-        logtrace(LogIntraPred,"%02x ", pred[x+y*stride]);
+        logtrace(LogIntraPred,"%02x ", dst[x+y*dstStride]);
 
       logtrace(LogIntraPred,"\n");
     }
@@ -693,21 +1006,17 @@ void intra_prediction_angular(de265_image* img,
 
 
 template <class pixel_t>
-void intra_prediction_planar(de265_image* img,int xB0,int yB0,int nT,int cIdx,
+void intra_prediction_planar(pixel_t* dst, int dstStride,
+                             int nT,int cIdx,
                              pixel_t* border)
 {
-  pixel_t* pred;
-  int      stride;
-  pred   = img->get_image_plane_at_pos_NEW<pixel_t>(cIdx,xB0,yB0);
-  stride = img->get_image_stride(cIdx);
-
   int Log2_nT = Log2(nT);
 
   for (int y=0;y<nT;y++)
     for (int x=0;x<nT;x++)
       {
-        pred[x+y*stride] = ((nT-1-x)*border[-1-y] + (x+1)*border[ 1+nT] +
-                            (nT-1-y)*border[ 1+x] + (y+1)*border[-1-nT] + nT) >> (Log2_nT+1);
+        dst[x+y*dstStride] = ((nT-1-x)*border[-1-y] + (x+1)*border[ 1+nT] +
+                              (nT-1-y)*border[ 1+x] + (y+1)*border[-1-nT] + nT) >> (Log2_nT+1);
       }
 
 
@@ -716,7 +1025,7 @@ void intra_prediction_planar(de265_image* img,int xB0,int yB0,int nT,int cIdx,
   for (int y=0;y<nT;y++)
     {
       for (int x=0;x<nT;x++)
-        logtrace(LogIntraPred,"%02x ", pred[x+y*stride]);
+        logtrace(LogIntraPred,"%02x ", dst[x+y*dstStride]);
 
       logtrace(LogIntraPred,"\n");
     }
@@ -724,14 +1033,10 @@ void intra_prediction_planar(de265_image* img,int xB0,int yB0,int nT,int cIdx,
 
 
 template <class pixel_t>
-void intra_prediction_DC(de265_image* img,int xB0,int yB0,int nT,int cIdx,
+void intra_prediction_DC(pixel_t* dst, int dstStride,
+                         int nT,int cIdx,
                          pixel_t* border)
 {
-  pixel_t* pred;
-  int      stride;
-  pred   = img->get_image_plane_at_pos_NEW<pixel_t>(cIdx,xB0,yB0);
-  stride = img->get_image_stride(cIdx);
-
   int Log2_nT = Log2(nT);
 
   int dcVal = 0;
@@ -745,32 +1050,22 @@ void intra_prediction_DC(de265_image* img,int xB0,int yB0,int nT,int cIdx,
   dcVal >>= Log2_nT+1;
 
   if (cIdx==0 && nT<32) {
-    pred[0] = (border[-1] + 2*dcVal + border[1] +2) >> 2;
+    dst[0] = (border[-1] + 2*dcVal + border[1] +2) >> 2;
 
-    for (int x=1;x<nT;x++) { pred[x]        = (border[ x+1] + 3*dcVal+2)>>2; }
-    for (int y=1;y<nT;y++) { pred[y*stride] = (border[-y-1] + 3*dcVal+2)>>2; }
+    for (int x=1;x<nT;x++) { dst[x]           = (border[ x+1] + 3*dcVal+2)>>2; }
+    for (int y=1;y<nT;y++) { dst[y*dstStride] = (border[-y-1] + 3*dcVal+2)>>2; }
     for (int y=1;y<nT;y++)
       for (int x=1;x<nT;x++)
         {
-          pred[x+y*stride] = dcVal;
+          dst[x+y*dstStride] = dcVal;
         }
   } else {
     for (int y=0;y<nT;y++)
       for (int x=0;x<nT;x++)
         {
-          pred[x+y*stride] = dcVal;
+          dst[x+y*dstStride] = dcVal;
         }
   }
-
-
-  logtrace(LogIntraPred,"INTRAPRED DC\n");
-  for (int y=0;y<nT;y++)
-    {
-      for (int x=0;x<nT;x++)
-        logtrace(LogIntraPred,"%02x ", pred[x+y*stride]);
-
-      logtrace(LogIntraPred,"\n");
-    }
 }
 
 
@@ -779,29 +1074,38 @@ template <class pixel_t>
 void decode_intra_prediction_internal(de265_image* img,
                                       int xB0,int yB0,
                                       enum IntraPredMode intraPredMode,
+                                      pixel_t* dst, int dstStride,
                                       int nT, int cIdx)
 {
-  pixel_t  border_pixels_mem[2*64+1];
-  pixel_t* border_pixels = &border_pixels_mem[64];
+  pixel_t  border_pixels_mem[4*MAX_INTRA_PRED_BLOCK_SIZE+1];
+  pixel_t* border_pixels = &border_pixels_mem[2*MAX_INTRA_PRED_BLOCK_SIZE];
 
   fill_border_samples(img, xB0,yB0, nT, cIdx, border_pixels);
 
-  if (img->sps.range_extension.intra_smoothing_disabled_flag == 0 &&
-      (cIdx==0 || img->sps.ChromaArrayType==CHROMA_444))
+  if (img->get_sps().range_extension.intra_smoothing_disabled_flag == 0 &&
+      (cIdx==0 || img->get_sps().ChromaArrayType==CHROMA_444))
     {
-      intra_prediction_sample_filtering(img, border_pixels, nT, cIdx, intraPredMode);
+      intra_prediction_sample_filtering(img->get_sps(), border_pixels, nT, cIdx, intraPredMode);
     }
 
 
   switch (intraPredMode) {
   case INTRA_PLANAR:
-    intra_prediction_planar(img,xB0,yB0,nT,cIdx, border_pixels);
+    intra_prediction_planar(dst,dstStride, nT,cIdx, border_pixels);
     break;
   case INTRA_DC:
-    intra_prediction_DC(img,xB0,yB0,nT,cIdx, border_pixels);
+    intra_prediction_DC(dst,dstStride, nT,cIdx, border_pixels);
     break;
   default:
-    intra_prediction_angular(img,xB0,yB0,intraPredMode,nT,cIdx, border_pixels);
+    {
+      int bit_depth = img->get_bit_depth(cIdx);
+      bool disableIntraBoundaryFilter =
+        (img->get_sps().range_extension.implicit_rdpcm_enabled_flag &&
+         img->get_cu_transquant_bypass(xB0,yB0));
+
+      intra_prediction_angular(dst,dstStride, bit_depth,disableIntraBoundaryFilter,
+                               xB0,yB0,intraPredMode,nT,cIdx, border_pixels);
+    }
     break;
   }
 }
@@ -821,9 +1125,111 @@ void decode_intra_prediction(de265_image* img,
   */
 
   if (img->high_bit_depth(cIdx)) {
-    decode_intra_prediction_internal<uint16_t>(img,xB0,yB0, intraPredMode,nT,cIdx);
+    decode_intra_prediction_internal<uint16_t>(img,xB0,yB0, intraPredMode,
+                                               img->get_image_plane_at_pos_NEW<uint16_t>(cIdx,xB0,yB0),
+                                               img->get_image_stride(cIdx),
+                                               nT,cIdx);
   }
   else {
-    decode_intra_prediction_internal<uint8_t>(img,xB0,yB0, intraPredMode,nT,cIdx);
+    decode_intra_prediction_internal<uint8_t>(img,xB0,yB0, intraPredMode,
+                                              img->get_image_plane_at_pos_NEW<uint8_t>(cIdx,xB0,yB0),
+                                              img->get_image_stride(cIdx),
+                                              nT,cIdx);
   }
 }
+
+
+// TODO: remove this
+template <> void decode_intra_prediction<uint8_t>(de265_image* img,
+                                                  int xB0,int yB0,
+                                                  enum IntraPredMode intraPredMode,
+                                                  uint8_t* dst, int nT, int cIdx)
+{
+    decode_intra_prediction_internal<uint8_t>(img,xB0,yB0, intraPredMode,
+                                              dst,nT,
+                                              nT,cIdx);
+}
+
+
+// TODO: remove this
+template <> void decode_intra_prediction<uint16_t>(de265_image* img,
+                                                   int xB0,int yB0,
+                                                   enum IntraPredMode intraPredMode,
+                                                   uint16_t* dst, int nT, int cIdx)
+{
+  decode_intra_prediction_internal<uint16_t>(img,xB0,yB0, intraPredMode,
+                                             dst,nT,
+                                             nT,cIdx);
+}
+
+
+template <class pixel_t>
+void decode_intra_prediction_from_tree_internal(const de265_image* img,
+                                                const enc_tb* tb,
+                                                const CTBTreeMatrix& ctbs,
+                                                const seq_parameter_set& sps,
+                                                int cIdx)
+{
+  enum IntraPredMode intraPredMode;
+  if (cIdx==0) intraPredMode = tb->intra_mode;
+  else         intraPredMode = tb->intra_mode_chroma;
+
+  pixel_t* dst = tb->intra_prediction[cIdx]->get_buffer<pixel_t>();
+  int dstStride = tb->intra_prediction[cIdx]->getStride();
+
+  pixel_t  border_pixels_mem[4*MAX_INTRA_PRED_BLOCK_SIZE+1];
+  pixel_t* border_pixels = &border_pixels_mem[2*MAX_INTRA_PRED_BLOCK_SIZE];
+
+  fill_border_samples_from_tree(img, tb, ctbs, cIdx, border_pixels);
+
+  if (cIdx==0) {
+    // memcpy(tb->debug_intra_border, border_pixels_mem, 2*64+1);
+  }
+
+  int nT = 1<<tb->log2Size;
+  if (cIdx>0 && tb->log2Size>2 && sps.chroma_format_idc == CHROMA_420) {
+    nT >>= 1; // TODO: 4:2:2
+  }
+
+
+  if (sps.range_extension.intra_smoothing_disabled_flag == 0 &&
+      (cIdx==0 || sps.ChromaArrayType==CHROMA_444))
+    {
+      intra_prediction_sample_filtering(sps, border_pixels, nT, cIdx, intraPredMode);
+    }
+
+
+  switch (intraPredMode) {
+  case INTRA_PLANAR:
+    intra_prediction_planar(dst,dstStride, nT, cIdx, border_pixels);
+    break;
+  case INTRA_DC:
+    intra_prediction_DC(dst,dstStride, nT, cIdx, border_pixels);
+    break;
+  default:
+    {
+      //int bit_depth = img->get_bit_depth(cIdx);
+      int bit_depth = 8; // TODO
+
+      bool disableIntraBoundaryFilter =
+        (sps.range_extension.implicit_rdpcm_enabled_flag &&
+         tb->cb->cu_transquant_bypass_flag);
+
+      intra_prediction_angular(dst,dstStride, bit_depth,disableIntraBoundaryFilter,
+                               tb->x,tb->y,intraPredMode,nT,cIdx, border_pixels);
+    }
+    break;
+  }
+}
+
+
+void decode_intra_prediction_from_tree(const de265_image* img,
+                                       const enc_tb* tb,
+                                       const CTBTreeMatrix& ctbs,
+                                       const seq_parameter_set& sps,
+                                       int cIdx)
+{
+  // TODO: high bit depths
+
+  decode_intra_prediction_from_tree_internal<uint8_t>(img ,tb, ctbs, sps, cIdx);
+}
