@@ -31,23 +31,31 @@ extern const int intraPredAngle_table[1+34];
    block (which is (x>>Log2MinPUSize) + (y>>Log2MinPUSize)*PicWidthInMinPUs).
    availableA/B is the output of check_CTB_available().
  */
-void fillIntraPredModeCandidates(int candModeList[3],
+void fillIntraPredModeCandidates(enum IntraPredMode candModeList[3],
                                  int x,int y, int PUidx,
                                  bool availableA, // left
                                  bool availableB, // top
                                  const de265_image* img);
 
 
-inline void fillIntraPredModeCandidates(int candModeList[3], int x,int y,
+inline void fillIntraPredModeCandidates(enum IntraPredMode candModeList[3], int x,int y,
                                  bool availableA, // left
                                  bool availableB, // top
                                  const de265_image* img)
 {
-  int PUidx = img->sps.getPUIndexRS(x,y);
+  int PUidx = img->get_sps().getPUIndexRS(x,y);
   fillIntraPredModeCandidates(candModeList, x,y, PUidx, availableA,availableB, img);
 }
 
 
+void fillIntraPredModeCandidates(enum IntraPredMode candModeList[3],
+                                 int x,int y,
+                                 bool availableA, // left
+                                 bool availableB, // top
+                                 const class CTBTreeMatrix& ctbs,
+                                 const seq_parameter_set* sps);
+
+
 
 /* Return value >= 0 -> use mpm_idx(return value)
    else              -> use rem_intra(-return value-1)
@@ -55,7 +63,7 @@ inline void fillIntraPredModeCandidates(int candModeList[3], int x,int y,
    This function may modify the candModeList !
  */
 int find_intra_pred_mode(enum IntraPredMode mode,
-                         int candModeList[3]);
+                         enum IntraPredMode candModeList[3]);
 
 void list_chroma_pred_candidates(enum IntraPredMode chroma_mode[5],
                                  enum IntraPredMode luma_mode);
@@ -69,6 +77,7 @@ int get_intra_scan_idx_chroma(int log2TrafoSize, enum IntraPredMode intraPredMod
 enum IntraPredMode lumaPredMode_to_chromaPredMode(enum IntraPredMode luma,
                                                   enum IntraChromaPredMode chroma);
 
+/*
 void decode_intra_block(decoder_context* ctx,
                         thread_context* tctx,
                         int cIdx,
@@ -77,6 +86,7 @@ void decode_intra_block(decoder_context* ctx,
                         int log2TrafoSize, int trafoDepth,
                         enum IntraPredMode intraPredMode,
                         bool transform_skip_flag);
+*/
 
 //void fill_border_samples(decoder_context* ctx, int xB,int yB,
 //                         int nT, int cIdx, uint8_t* out_border);
@@ -86,4 +96,16 @@ void decode_intra_prediction(de265_image* img,
                              enum IntraPredMode intraPredMode,
                              int nT, int cIdx);
 
+void decode_intra_prediction_from_tree(const de265_image* img,
+                                       const class enc_tb* tb,
+                                       const class CTBTreeMatrix& ctbs,
+                                       const class seq_parameter_set& sps,
+                                       int cIdx);
+
+// TODO: remove this
+template <class pixel_t> void decode_intra_prediction(de265_image* img,
+                                                      int xB0,int yB0,
+                                                      enum IntraPredMode intraPredMode,
+                                                      pixel_t* dst, int nT, int cIdx);
+
 #endif
