@@ -27,7 +27,7 @@
 #include <string.h>
 #if defined(_MSC_VER) || defined(__MINGW32__)
 # include <malloc.h>
-#else
+#elif defined(HAVE_ALLOCA_H)
 # include <alloca.h>
 #endif
 
@@ -191,6 +191,7 @@ pic_parameter_set::~pic_parameter_set()
 void pic_parameter_set::set_defaults(enum PresetSet)
 {
   pps_read = false;
+  sps = NULL;
 
   pic_parameter_set_id = 0;
   seq_parameter_set_id = 0;
@@ -306,12 +307,13 @@ bool pic_parameter_set::read(bitreader* br, decoder_context* ctx)
   num_ref_idx_l1_default_active++;
 
 
-  seq_parameter_set* sps = ctx->get_sps(seq_parameter_set_id);
-  if (sps->sps_read==false) {
+  if (!ctx->has_sps(seq_parameter_set_id)) {
     ctx->add_warning(DE265_WARNING_NONEXISTING_SPS_REFERENCED, false);
     return false;
   }
 
+  sps = ctx->get_sps(seq_parameter_set_id);
+
   if ((pic_init_qp = get_svlc(br)) == UVLC_ERROR) {
     ctx->add_warning(DE265_WARNING_PPS_HEADER_INVALID, false);
     return false;
